/*
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*global window, document, setTimeout, setInterval, clearInterval,
 requestAnimationFrame */

window.app = window.app || {};

/**
 * App module.
 *
 * @requires (@link app.systeminfo}
 * @module app
 * @namespace app
 */

(function defineApp(app) {
    'use strict';

    /**
     * Variable which will be used to
     * store Puzzle object.
     *
     * @memberof app
     * @private
     * @type {object}
     */
    var game = null;

    /**
     * Creates instance of Puzzle object.
     *
     * @namespace Puzzle
     * @memberof app
     * @public
     * @returns {object}
     */
    function Puzzle() {

        /**
         * Duration in milliseconds of changing piece position
         * performed during shuffling pieces.
         *
         * @memberof Puzzle
         * @private
         * @const {number}
         */
        var SHUFFLE_TIME = 250,

            /**
             * Duration in milliseconds of piece animation performed
             * during final animation.
             *
             * @memberof Puzzle
             * @private
             * @const {number}
             */
            SHORT_DELAY = 60,

            /**
             * Duration in milliseconds of piece animation performed
             * during shuffling animation.
             *
             * @memberof Puzzle
             * @private
             * @const {number}
             */
            LONG_DELAY = 100,

            /**
             * When user starts holding the screen and does not stop it
             * for this time in milliseconds, the game is restarted.
             *
             * @memberof Puzzle
             * @private
             * @const {number}
             */
            HOLD_TIME = 2000,

            /**
             * Time of piece movement animation in milliseconds.
             *
             * @memberof Puzzle
             * @private
             * @const {number}
             */
            ANIMATION_TIME = 400,

            /**
             * Identifier of interval action that changes piece's
             * position during shuffling pieces.
             *
             * @memberof Puzzle
             * @private
             * @type {number}
             */
            interval = 0,

            /**
             * Indicates if detected touch event is single touch or not.
             *
             * @memberof Puzzle
             * @private
             * @type {boolean}
             */
            singleTouch = false,

            /**
             * Index of last randomly chosen background.
             *
             * @memberof Puzzle
             * @private
             * @type {number}
             */
            lastRandom = -1,

            /**
             * Names of available background images.
             *
             * @memberof Puzzle
             * @private
             * @type {string[]}
             */
            images = ['1.jpg', '2.jpg', '3.jpg', '4.jpg', '5.jpg'],

            /**
             * Path of the current background image.
             *
             * @memberof Puzzle
             * @private
             * @type {string}
             */
            file = '',

            /**
             * Resolution of the screen.
             *
             * @memberof Puzzle
             * @private
             * @type {object}
             * @property {number} width
             * @property {number} height
             */
            resolution = {
                width: window.innerWidth || 320,
                height: window.innerHeight || 320
            },

            /**
             * Dimensions of the board,
             * x maximum count of pieces on horizontal axis,
             * y Maximum count of pieces on vertical axis.
             *
             * @memberof Puzzle
             * @private
             * @type {object}
             * @property {number} x
             * @property {number} y
             */
            grid = {
                x: 4,
                y: 4
            },

            /**
             * Width of a piece.
             *
             * @memberof Puzzle
             * @private
             * @type {number}
             */
            pieceWidth = 0,

            /**
             * Height of a piece.
             *
             * @memberof Puzzle
             * @private
             * @type {number}
             */
            pieceHeight = 0,

            /**
             * Position of free space on the board.
             *
             * @memberof Puzzle
             * @private
             * @type {object}
             * @property {number} left
             * @property {number} top
             */
            free = {left: null, top: null},

            /**
             * Last moved piece.
             *
             * @memberof Puzzle
             * @private
             * @type {object}
             */
            lastPiece = null,

            /**
             * Indicates if user can move pieces or not.
             *
             * @memberof Puzzle
             * @private
             * @type {boolean}
             */
            lock = false,

            /**
             * Reference to background element.
             *
             * @memberof Puzzle
             * @private
             * @type {HTMLElement}
             */
            backgroundEl = document.getElementById('background'),

            /**
             * Reference to semitransparent background element,
             * that makes border of game board on circular screens.
             *
             * @memberof Puzzle
             * @private
             * @type {HTMLElement}
             */
            transparentEl = document.getElementById('transparent'),

            /**
             * Reference to game board element.
             *
             * @memberof Puzzle
             * @private
             * @type {HTMLElement}
             */
            gameEl = document.getElementById('game'),

            /**
             * Offset of game board element (horizontally and vertically,
             * both takes the same value).
             *
             * @memberof Puzzle
             * @private
             * @type {number}
             */
            gameElOffset = 0,

            /**
             * Indicates if the screen is circular or not.
             *
             * @memberof Puzzle
             * @private
             * @type {boolean}
             */
            isScreenCircular = window.matchMedia(
                '(-tizen-geometric-shape: circle)'
            ).matches;

        /**
         * Returns random background image name.
         *
         * @memberof Puzzle
         * @private
         * @returns {string}
         */
        function randomFile() {
            var random = Math.floor(Math.random() * images.length);

            if (images.length === 1) {
                return images[0];
            }

            if (random === lastRandom) {
                return randomFile();
            }

            lastRandom = random;
            return images[random];
        }

        /**
         * Removes all pieces from the board.
         *
         * @memberof Puzzle
         * @private
         */
        function clearGame() {
            lock = false;
            clearInterval(interval);
            var pieces = document.getElementsByClassName('piece'),
                i = pieces.length;

            while (i) {
                i -= 1;
                pieces[i].remove();
            }
        }

        /**
         * Returns position in pixels for piece coordinates.
         *
         * @memberof Puzzle
         * @private
         * @param {number} x
         * @param {number} y
         * @returns {{left: number, top: number}}
         */
        function getPosition(x, y) {
            return {
                left: (x * pieceWidth),
                top: (y * pieceHeight)
            };
        }

        /**
         * Sets random file.
         *
         * @memberof Puzzle
         * @private
         */
        function setFile() {
            file = 'images/' + resolution.width + 'x' +
                resolution.height + '/' +
                randomFile();
        }

        /**
         * Animates an element.
         *
         * @memberof Puzzle
         * @private
         * @param {HTMLElement} el
         * @param {string} propertyName
         * @param {number} targetValue
         * @param {string} unit
         */
        function animateElement(el, propertyName, targetValue, unit) {
            var animationStartTime = +new Date(),
                startValue = parseFloat(el.style[propertyName], 10),
                diff = targetValue - startValue,
                tick;

            el.style[propertyName] = startValue + unit;

            tick = function tick() {
                var // time passed since the start (in milliseconds)
                    currentAnimationTime = new Date() - animationStartTime,
                // current progress (amount of animation which should be
                // completed)
                    currentProgress = currentAnimationTime / ANIMATION_TIME;

                if (currentProgress >= 1) {
                    el.style[propertyName] = targetValue + unit;
                    return true;
                }

                el.style[propertyName] = (
                        startValue + currentProgress * diff
                    ) + unit;

                requestAnimationFrame(tick);
                return false;
            };

            tick();
        }

        /**
         * Sets background image.
         *
         * @memberof Puzzle
         * @private
         */
        function setBackground() {
            backgroundEl.style.backgroundImage = 'url(' + file + ')';
            backgroundEl.style.display = 'block';
            backgroundEl.style.opacity = 1;
            animateElement(backgroundEl, 'opacity', 0.4, '');
            backgroundEl.style.webkitFilter = 'invert(100%)';
            transparentEl.style.backgroundImage = 'url(' + file + ')';
        }

        /**
         * Returns random piece adjacent to the free space.
         *
         * @memberof Puzzle
         * @private
         * @returns {HTMLElement}
         */
        function pickRandomNearest() {
            var nearest = [],
                nearestLen = 0,
                pieces = [],
                i = 0,
                last = Infinity,
                current = 0,

            // calculate the central point of the free tile
                position = getPosition(free.x, free.y),
                left = gameElOffset + position.left + (pieceWidth / 2),
                top = gameElOffset + position.top + (pieceHeight / 2);

            // Find pieces adjacent to the free tile
            nearest.push(document.elementFromPoint(left, top - pieceHeight));
            nearest.push(document.elementFromPoint(left, top + pieceHeight));
            nearest.push(document.elementFromPoint(left - pieceWidth, top));
            nearest.push(document.elementFromPoint(left + pieceWidth, top));

            // remove empty values
            nearest = nearest.filter(function notEmpty(value) {
                return value;
            });

            nearestLen = nearest.length;
            for (i; i < nearestLen; i += 1) {
                current = nearest[i].step;
                if (nearest[i].classList.contains('piece') &&
                    nearest[i] !== lastPiece &&
                    Math.min(last, current) === current &&
                    (nearest[i].left !== free.left ||
                    nearest[i].top !== free.top)) {
                    pieces.push(nearest[i]);
                    last = nearest[i].step;
                }
            }

            lastPiece = pieces[Math.floor(Math.random() * pieces.length)];

            return lastPiece;
        }

        /**
         * Checks if a piece with given coordinates can be moved.
         * Returns true if can be moved, false otherwise.
         *
         * @memberof Puzzle
         * @private
         * @param {number} x
         * @param {number} y
         * @returns {boolean}
         */
        function canMove(x, y) {
            var leftDiff = Math.abs(x - free.x),
                topDiff = Math.abs(y - free.y);

            return (leftDiff === 0 && topDiff === 1) ||
                (leftDiff === 1 && topDiff === 0);
        }

        /**
         * Sets free tile position.
         *
         * @memberof Puzzle
         * @private
         * @param {number} x
         * @param {number} y
         */
        function setFree(x, y) {
            free.x = x;
            free.y = y;
        }

        /**
         * Checks if the piece position is correct.
         *
         * @memberof Puzzle
         * @private
         * @param {HTMLElement} piece
         */
        function matchPosition(piece) {
            if (piece.x === piece.startX &&
                piece.y === piece.startY) {
                piece.classList.add('match');

                setTimeout(function animate() {
                    var position = getPosition(piece.x, piece.y);
                    // don't blink if the piece was already moved away
                    if (position.left !== parseInt(piece.style.left, 10) ||
                        position.top !== parseInt(piece.style.top, 10)) {
                        return;
                    }
                    piece.style.opacity = 0;
                    animateElement(piece, 'opacity', 1, '');
                }, ANIMATION_TIME + 50); // must wait for the animation to end
            }
        }

        /**
         * Increments step count for the piece.
         *
         * @memberof Puzzle
         * @private
         * @param {HTMLElement} piece
         */
        function setStepCount(piece) {
            piece.step += 1;
        }

        /**
         * Creates the board.
         *
         * @memberof Puzzle
         * @private
         */
        function createPieces() {
            var i = 0,
                j = 0,
                piece,
                position;

            for (i; i < grid.y; i += 1) {
                for (j = 0; j < grid.x; j += 1) {
                    if (i === (grid.y - 1) && j === (grid.x - 1)) {
                        // leave the last piece empty
                        break;
                    }

                    piece = document.createElement('div');
                    piece.className = 'piece match';
                    piece.x = j;
                    piece.y = i;
                    piece.startX = j;
                    piece.startY = i;
                    piece.number = i * grid.x + j;
                    piece.setAttribute('id', 'piece_' + j + '_' + i);
                    piece.step = 0;
                    piece.style.width = (pieceWidth - 1) + 'px';
                    piece.style.height = (pieceHeight - 1) + 'px';

                    position = getPosition(piece.x, piece.y);
                    piece.style.left = position.left + 'px';
                    piece.style.top = position.top + 'px';
                    piece.style.position = 'absolute';
                    piece.style.opacity = 1;
                    piece.style.backgroundImage = 'url(' + file + ')';
                    piece.style.backgroundPosition =
                        '-' + (gameElOffset + piece.x *
                        (pieceWidth - 1) + j + 1) +
                        'px -' +
                        (gameElOffset + piece.y * (pieceHeight - 1) +
                        i + 1) + 'px';
                    document.getElementById('pieces').appendChild(piece);
                }
            }
        }


        /**
         * Does the final fade out piece animation.
         * Parameter last is the flag that indicates
         * if animated piece is last one.
         *
         * @memberof Puzzle
         * @private
         * @param {HTMLElement} piece
         * @param {boolean} last
         */
        function finalPieceAnimation(piece, last) {
            animateElement(piece, 'width', pieceWidth * 2, 'px');
            animateElement(piece, 'height', pieceHeight * 2, 'px');
            animateElement(piece, 'left',
                parseInt(piece.style.left, 10) - (pieceWidth / 2), 'px'
            );
            animateElement(piece, 'top',
                parseInt(piece.style.top, 10) - (pieceHeight / 2), 'px'
            );
            animateElement(piece, 'opacity', 0, '');

            if (last) {
                window.addEventListener('click', function restart() {
                    window.removeEventListener('click', restart);
                    game.restart();
                });
            }
        }

        /**
         * Schedules the final animation of the piece with given delay.
         * Parameter last is flag that indicates if animated piece is last one.
         *
         * @memberof Puzzle
         * @private
         * @param {HTMLElement} piece
         * @param {number} delay
         * @param {boolean} last
         * @returns {number}
         */
        function scheduleFinalPieceAnimation(piece, delay, last) {
            return setTimeout(function makeFinalPieceAnimation() {
                finalPieceAnimation(piece, last);
            }, delay);
        }

        /**
         * Does the final fade out animation of the board.
         *
         * @memberof Puzzle
         * @private
         */
        function finalAnimation() {
            var pieces = document.getElementsByClassName('piece'),
                piecesLen = pieces.length;

            backgroundEl.style.display = 'none';
            transparentEl.classList.add('opaque');

            while (piecesLen) {
                piecesLen -= 1;
                scheduleFinalPieceAnimation(
                    pieces[piecesLen],
                    piecesLen * SHORT_DELAY,
                    piecesLen === 0
                );
            }
        }

        /**
         * Checks the winning condition (all pieces matched).
         * Triggers final fade out animation of the board.
         * Sets background opacity to 1.
         *
         * @memberof Puzzle
         * @private
         */
        function checkWin() {
            if (document.querySelectorAll('.piece:not(.match)').length === 0) {
                finalAnimation();
                backgroundEl.style.webkitFilter = 'none';
                animateElement(backgroundEl, 'opacity', 1, '');
            }
        }

        /**
         * Handles piece position change event.
         * Checks if the piece position is correct.
         * Increments step count for the piece.
         * Sets free tile position.
         * Checks the winning condition.
         *
         * @memberof Puzzle
         * @private
         * @param {HTMLElement} piece
         * @param {number} left
         * @param {number} top
         */
        function onChangePiecePosition(piece, left, top) {
            matchPosition(piece);
            setStepCount(piece);
            setFree(left, top);
            checkWin();
            lock = false;
        }

        /**
         * Moves a random piece into the free space.
         *
         * @memberof Puzzle
         * @private
         * @param {HTMLElement} piece
         */
        function changePiecePosition(piece) {
            var x = 0,
                y = 0,
                position = null;

            piece = piece || pickRandomNearest();
            x = piece.x;
            y = piece.y;
            if (!lock && canMove(x, y)) {
                lock = true;
                piece.classList.toggle('match', false);

                position = getPosition(free.x, free.y);

                // the piece goes into the free space
                piece.x = free.x;
                piece.y = free.y;
                animateElement(piece, 'left', position.left, 'px');
                animateElement(piece, 'top', position.top, 'px');

                onChangePiecePosition(piece, x, y);
            }
        }

        /**
         * Randomly shuffles all pieces.
         *
         * @memberof Puzzle
         * @private
         */
        function shufflePieces() {
            interval = setInterval(changePiecePosition, LONG_DELAY);
            setTimeout(function stopChangingPiecePosition() {
                clearInterval(interval);
                interval = null;
            }, grid.x * grid.y * SHUFFLE_TIME);
        }

        /**
         * Starts the game.
         * If this is the first start of the game, initializes
         * systeminfo module.
         *
         * @memberof Puzzle
         * @public
         * @param {boolean} [firstStart=false]
         *
         */
        function start(firstStart) {
            transparentEl.classList.remove('opaque');
            clearGame();
            setFile();
            setBackground();
            setFree(grid.x - 1, grid.y - 1);
            createPieces();
            shufflePieces();
            if (firstStart === true) {
                app.systeminfo.init();
            }
        }

        /**
         * Handles if the hold event time is long enough to restart the game.
         *
         * @memberof Puzzle
         * @private
         * @param {number} startTime
         */
        function onHoldEnd(startTime) {
            var endTime = new Date().getTime();

            if ((endTime - startTime) > HOLD_TIME) {
                if (interval !== null || singleTouch === true) {
                    // do not restart if shuffling in progress
                    // single touch hold shouldn't cause restart (2nd condition)
                    return;
                }
                start();
            }
        }

        /**
         * Handles hold event.
         *
         * @memberof Puzzle
         * @private
         * @param {TouchEvent} event
         */
        function onHoldStart(event) {
            var startTime = new Date().getTime(),
                touchEnd = function touchEnd() {
                    gameEl.removeEventListener('touchend', touchEnd);
                    onHoldEnd(startTime);
                },
                touches = event.touches.length;

            // Prevent reset while single touch hold.
            singleTouch = touches < 2;

            gameEl.addEventListener('touchend', touchEnd);
        }

        /**
         * Binds events.
         *
         * @memberof Puzzle
         * @private
         */
        function bindEvents() {
            gameEl.addEventListener('touchstart', onHoldStart);
        }

        /**
         * Calculates size of pieces and updates game board element's height for
         * rectangular screen.
         *
         * @memberof Puzzle
         * @private
         */
        function adjustGameElementForRectangularScreen() {
            gameEl.style.height = window.innerHeight + 'px';
            pieceWidth = Math.floor(resolution.width / grid.x);
            pieceHeight = Math.floor(resolution.height / grid.y);
        }

        /**
         * Calculates size of pieces and updates game board element's size and
         * offset for circular screen.
         *
         * @memberof Puzzle
         * @private
         */
        function adjustGameElementForCircularScreen() {
            var sideLength = Math.floor(window.innerHeight / Math.sqrt(2));
            gameElOffset = Math.floor((window.innerWidth - sideLength) / 2);

            pieceWidth = Math.floor(sideLength / grid.x);
            pieceHeight = Math.floor(sideLength / grid.y);

            gameEl.style.width = pieceWidth * grid.x + 1 + 'px';
            gameEl.style.height = pieceHeight * grid.y + 1 + 'px';
            gameEl.style.marginLeft = gameElOffset + 'px';
            gameEl.style.marginTop = gameElOffset + 'px';
        }

        /**
         * Sets up game board's size and position.
         *
         * @memberof Puzzle
         * @private
         */
        function adjustGameElement() {
            if (isScreenCircular) {
                adjustGameElementForCircularScreen();
            } else {
                adjustGameElementForRectangularScreen();
            }
        }

        /**
         * Initializes the game and binds events.
         *
         * @memberof Puzzle
         * @private
         */
        function init() {
            adjustGameElement();
            bindEvents();

            // add eventListener for tizenhwkey
            document.addEventListener('tizenhwkey', function onTizenhwkey(e) {
                if (e.keyName === 'back' && !!app.systeminfo) {
                    app.systeminfo.exit();
                }
            });

            document.getElementById('pieces').addEventListener('click',
                function movePiece(event) {
                    if (event.target.classList.contains('piece')) {
                        changePiecePosition(event.target);
                    }
                });
        }

        init();

        return {
            start: start,
            restart: start
        };
    }
    // Creates game object
    game = new Puzzle();
    game.start(true);
})(window.app);
